/*************************************************
*     Exim - an Internet mail transport agent    *
*************************************************/

/* Copyright (c) University of Cambridge 1995 - 2006 */
/* See the file NOTICE for conditions of use and distribution. */

/*---------------------------------------------------------------------------
* Copyright 2011 Igor Daniloff.
*
* Following source code is the property of Doctor Web, Ltd.
* Use is subject to license terms.
* The reproduction in any form without written permission of
* Doctor Web, Ltd. and proper attribution is strongly prohibited.
----------------------------------------------------------------------------*/

#include "local_scan.h"

#include "DwRsSDK_Log.H"
#include "DwRsSDK_IReceiver.H"
#include "DwRsSDK_Mail.H"
#include "DwRsSDK_IComponentOptions.H"

#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

enum Consts{
    LOG_BUF_LEN = 512,
    IO_BUF_LEN  = 4096,
    INVALID_FILE_DESC=-1,
    LOCAL_SCAN_DISCARD=777
};


static int g_timeout = 30;
static uschar *g_base_dir = US"/var/drweb/";
static int g_res_on_error = LOCAL_SCAN_TEMPREJECT;
static int g_max_size = 0;
static uschar *g_res_on_error_param = US"tempfail";
static uschar *g_log_level = US"info";
static uschar *g_ipc_log_level = US"alert";
static uschar *g_syslog_facility = US"mail";

optionlist local_scan_options[] = {/* in alphabetical order !!! */
    { "DrwebBaseDir",  opt_stringptr, &g_base_dir },
    { "DrwebIpcLogLevel",  opt_stringptr, &g_ipc_log_level },
    { "DrwebLogLevel",  opt_stringptr, &g_log_level },
    { "DrwebMaxSize",  opt_int, &g_max_size },
    { "DrwebProcessingError",  opt_stringptr, &g_res_on_error_param },
    { "DrwebSyslogFacility",  opt_stringptr, &g_syslog_facility },
    { "DrwebTimeout", opt_time,       &g_timeout }
};
int local_scan_options_count = sizeof(local_scan_options)/sizeof(optionlist);

void init_params()
{
    if(!strcasecmp(g_res_on_error_param,"tempfail"))
        g_res_on_error = LOCAL_SCAN_TEMPREJECT;
    else if(!strcasecmp(g_res_on_error_param,"reject"))
        g_res_on_error = LOCAL_SCAN_REJECT;
    else if(!strcasecmp(g_res_on_error_param,"discard"))
        g_res_on_error = LOCAL_SCAN_DISCARD;
    else if(!strcasecmp(g_res_on_error_param,"pass"))
        g_res_on_error = LOCAL_SCAN_ACCEPT;
    if(g_max_size<0)
        g_max_size=0;
}

/* ---------------------------------------------------------------------- */
int dw_write_buf( int to_fd, const char* buf, off_t buf_len )
{
    off_t sendbytes, written;

    sendbytes = 0;
    while( buf_len > 0 )
    {
        written = write( to_fd, buf + sendbytes, buf_len );
        if( written < 0  )
        {
            DwLog( LOG_ERR, "files: write buf %p[%d] to fd (%d) error - (%d) %s", buf, buf_len, to_fd, errno, strerror(errno) );
            return FALSE;
        }
        sendbytes += written;
        buf_len -= written;
    }
    return TRUE;
}

/* ---------------------------------------------------------------------- */
int dw_write_fd_to_fd( int from_fd, int to_fd )
{
    int          readed;
    char         buf[IO_BUF_LEN];

    while( 1 )
    {
        readed = read(from_fd,buf,IO_BUF_LEN);
        DwLog( LOG_DEBUG, "files: read = %i", readed );
        if( readed < 0 )
        {
            DwLog( LOG_ERR, "files: cannot read file with message" );
            return FALSE;
        }
        if( readed == 0 )
        {
            /* end-of-file */
            return TRUE;
        }
        if( dw_write_buf( to_fd, buf, readed )==FALSE)
        {
            return FALSE;
        }
    }
    return TRUE;
}

/* ---------------------------------------------------------------------- */
uschar* RemoveBrackets(uschar* str)
{
    if (str[0]=='<') {
        const size_t len = strlen(++str);
        if (len>0 && str[len-1]=='>') {
            str[len-1] = '\0';
        }
    }
    return str;
}

/* ---------------------------------------------------------------------- */
int UpdateEnvelope(dwrssdk_mail_t mail)
{
    unsigned int num=0;
    if(!DwRsSDK_LoadEnvelope(mail)){
        DwLog(LOG_ERR,"Can not read new envelope");
        return FALSE;
    }
    recipients_count=0;
    for(;num<DwRsSDK_GetRecipientsNum(mail);++num){
        uschar* new_rcpt = RemoveBrackets(string_copy(US(DwRsSDK_GetRecipientAt(mail,num))));
        receive_add_recipient(new_rcpt,-1);
    }
    return TRUE;
}
/* ---------------------------------------------------------------------- */
int UpdateHeaders(dwrssdk_mail_t mail)
{
    const char* name,*value;
    header_line * h_line;
    unsigned int num=0,i;
    if(!DwRsSDK_LoadHeaders(mail)){
        DwLog(LOG_DEBUG,"Can not read new headers");
        return TRUE;
    }
    h_line=header_list;
    while ( h_line != NULL && num<DwRsSDK_GetHeadersNum(mail)){
        if(h_line->type != '*'){
            uschar* delim=strchr(h_line->text,':');
            if(!DwRsSDK_GetHeaderAt(mail,num,&name,&value)){
                DwLog(LOG_ERR,"Error in add new header");
                return FALSE;
            }
            if(header_testname(h_line,US(name),strlen(name),TRUE)==FALSE||
               delim==NULL||
               delim==(h_line->text+strlen(h_line->text)-1)||
               strncmp(delim+2,value,strlen(value))
               ){
                DwLog(LOG_DEBUG,"Header '%s' changed. New header: '%s: %s'",h_line->text,name,value);
                break;
            }
            ++num;
        }
        if(h_line==header_last){
            h_line=0;
            break;
        }
        h_line = h_line->next;
    }
    while ( h_line != NULL ){
        h_line->type = '*';
        DwLog(LOG_DEBUG,"Remove header '%s'",h_line->text);
        if(h_line==header_last)
            break;
        if(h_line==header_last)
            break;
        h_line = h_line->next;
    }
    for(i=num;i<DwRsSDK_GetHeadersNum(mail);++i){
        if(!DwRsSDK_GetHeaderAt(mail,i,&name,&value)){
            DwLog(LOG_ERR,"Error in add new header");
            return FALSE;
        }
        header_add(' ',"%s: %s\n",name,value);
        DwLog(LOG_DEBUG,"Add header '%s: %s'",name,value);
    }
    return TRUE;
}

/*----------------------------------------------------------------------*/
int
local_scan(int fd, uschar **return_text)
{
    LogInfo log_info={"drweb_ls",
                      g_log_level,
                      g_syslog_facility,
                      0,
                      g_ipc_log_level,
                      ""
    };
    dwrssdk_mail_t mail=NULL;
    dwrssdk_receiver_t sender=NULL;
    dwrssdk_processing_info_t processing_info=NULL;
    recipient_item * item=NULL;
    int to_file_desc=INVALID_FILE_DESC;
    header_line * h_line;
    int action, res;
    int need_remove=FALSE;
    dwrssdk_component_options_t component_options=NULL;

    init_params();

    res=g_res_on_error;

    if(!InitLog(&log_info)){
        DwLog(LOG_ERR,"Error in init DwRsSDK log");
        goto exit;
    }

    if(!strcmp("drweb-scanned",received_protocol)){
        DwLog(LOG_DEBUG,"Msg is from Sender module -> skip it");
        res=LOCAL_SCAN_ACCEPT;
        goto exit;
    }

    if(g_max_size>0){
        struct stat st;
        if(fstat(fd,&st)!=-1&&st.st_size>g_max_size){
            DwLog(LOG_DEBUG,"Msg with size %d is accepted due size limit (max size=%d)",st.st_size,g_max_size);
            res=LOCAL_SCAN_ACCEPT;
            goto exit;
        }
    }

    if(!(sender=DwRsSDK_CreateReceiver())){
        DwLog(LOG_ERR,"Error in create Receiver object");
        goto exit;
    }
    if(!DwRsSDK_InitializeReceiver(sender,
                                   g_base_dir,
                                   0,
                                   kMO_ReadyForBlockAndSend | kMO_CanNotBodyModify,
                                   g_timeout,
                                   1,0,0,0,0,0)){
        DwLog(LOG_ERR,"Error on initialize Receiver object");
        goto exit;
    }
    if(!(processing_info=DwRsSDK_SetNewMsgDir(sender))){
        DwLog(LOG_ERR,"Can not create new msg dir for send!");
        goto exit;
    }
    mail = DwRsSDK_CreateMail(DwRsSDK_GetDirToSave(processing_info));
    if(!mail ||
       !DwRsSDK_SetSender(mail,sender_address)){
        DwLog(LOG_ERR,"Can not create Mail object!");
        goto exit;
    }
    for(item=recipients_list;item!=recipients_list+recipients_count;++item){
        if(!DwRsSDK_AddRecipient(mail,item->address)){
            DwLog(LOG_ERR,"Can not create Mail object (adding receipient error)!");
            goto exit;
        }
    }
    DwRsSDK_LogEnvelope(mail);
    if(!DwRsSDK_SaveEnvelope(mail)){
        DwLog(LOG_ERR,"Error in save envelope file to %s",
                 DwRsSDK_GetDirToSave(processing_info));
        goto exit;
    }
    to_file_desc=open(DwRsSDK_GetBodyFilename(mail),O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
    if(to_file_desc==INVALID_FILE_DESC){
        DwLog(LOG_ERR,"Error in open msg file %s for writing: %s",
                 DwRsSDK_GetBodyFilename(mail),strerror(errno));
        goto exit;
    }
    h_line=header_list;
    while ( h_line != NULL ){
        if(h_line->type != '*'){
            DwLog(LOG_DEBUG,"Write header type=%c, len=%d '%s'",h_line->type,h_line->slen,h_line->text);
            if(dw_write_buf(to_file_desc,h_line->text,h_line->slen)==FALSE){
                DwLog(LOG_ERR,"Error in header to field %s: %s",
                         DwRsSDK_GetBodyFilename(mail),strerror(errno));
                close(to_file_desc);
                goto exit;
            }
        }
        if(h_line==header_last)
            break;
        h_line = h_line->next;
    }
    if( dw_write_buf(to_file_desc,"\n",1)==FALSE ||
        dw_write_fd_to_fd(fd,to_file_desc)==FALSE){
        DwLog(LOG_ERR,"Error in save msg file to %s",
                 DwRsSDK_GetBodyFilename(mail));
        close(to_file_desc);
        goto exit;
    }
    close(to_file_desc);

    component_options = DwRsSDK_CreateComponentOptions(0,0);

    if(component_options){

        if(interface_address){
            char port[16];

            DwRsSDK_ComponentOptions_Add(component_options,DWRSSDK_CO_ServerIP,interface_address);

            snprintf(port,15,"%i",interface_port);
            DwRsSDK_ComponentOptions_Add(component_options,DWRSSDK_CO_ServerPort,port);
        }

        if(sender_host_address){
            DwRsSDK_ComponentOptions_Add(component_options,DWRSSDK_CO_ClientIP,sender_host_address);
        }

        if(sender_host_authenticated){
            DwRsSDK_ComponentOptions_Add(component_options,DWRSSDK_CO_Auth,sender_host_authenticated);
        }

        DwRsSDK_ComponentOptions_Add(component_options,DWRSSDK_CO_MsgId,message_id);
    }


    DwLog(LOG_INFO,"success prepare msg to dwreb-maild with number %s. Send request...",DwRsSDK_GetDataToSend(processing_info));

    action = DwRsSDK_SendRequestToMaild(sender,processing_info,component_options,0);

    if(action&kACT_Error){
        DwLog(LOG_ERR,"Error in check mail msg '%s' with Dr.Web Maild",DwRsSDK_GetDataToSend(processing_info));
        need_remove=TRUE;
        goto exit;
    }else if(action&kACT_Pass){
        need_remove=TRUE;
        if(action&kACT_WasBodyModify){
            DwLog(LOG_ERR,"Received body update, but can not support it!");
            goto exit;
        }
        if(action&kACT_WasEnvelopeChange){
            DwLog(LOG_DEBUG,"Envelope was changed: need update...");
            if(UpdateEnvelope(mail)==FALSE)
                goto exit;
        }
        if(action&kACT_WasTopHeaderModify){
            DwLog(LOG_DEBUG,"Headers was changed: need update...");
            if(UpdateHeaders(mail)==FALSE)
                goto exit;
        }
        res=LOCAL_SCAN_ACCEPT;
    }else if(action&kACT_Success){
        DwLog(LOG_INFO,"Msg '%s' is passed or will be checked later",DwRsSDK_GetDataToSend(processing_info));
        res=LOCAL_SCAN_DISCARD;
    } else if (action&kACT_Reject) {
        DwLog(LOG_INFO,"Msg '%s' is rejected. (smtp answer='%s')",DwRsSDK_GetDataToSend(processing_info),DwRsSDK_GetSmtpAnswer(processing_info));
        res=LOCAL_SCAN_REJECT;
    } else if (action&kACT_Discard) {
        DwLog(LOG_INFO,"Msg '%s' is discard. (smtp answer='%s')",DwRsSDK_GetDataToSend(processing_info),DwRsSDK_GetSmtpAnswer(processing_info));
        res=LOCAL_SCAN_DISCARD;
    } else if (action&kACT_Tempfail) {
        DwLog(LOG_INFO,"Msg '%s' is tempfail. (smtp answer='%s')",DwRsSDK_GetDataToSend(processing_info),DwRsSDK_GetSmtpAnswer(processing_info));
        res=LOCAL_SCAN_TEMPREJECT;
    } else {
        DwLog(LOG_ERR,"Unrecognize action from maild: 0x%x",action);
        goto exit;
    }
    if(strlen(DwRsSDK_GetSmtpAnswer(processing_info))>0){
        *return_text=string_copy(US(DwRsSDK_GetSmtpAnswer(processing_info)));
    }
    if(need_remove==FALSE){
        DwRsSDK_DeleteProcessingInfo(processing_info);
        processing_info=NULL;
    }
 exit:
    if(component_options)
        DwRsSDK_DeleteComponentOptions(component_options);
    if(mail)
        DwRsSDK_DeleteMail(mail);
    if(processing_info){
        DwRsSDK_RemoveDirWithInfo(processing_info);
        DwRsSDK_DeleteProcessingInfo(processing_info);
    }
    if(sender){
        DwRsSDK_DeleteReceiver(sender);
    }
    if(res==LOCAL_SCAN_DISCARD){
        recipients_count=0;
        res=LOCAL_SCAN_ACCEPT;
    }
    return res;

}

